"""Data models for ECS CNAME."""

from dataclasses import dataclass, fields
from typing import Dict, List


@dataclass(init=False, order=True)
class ECSNetworkMetadata:
    """ECS container network metadata."""

    NetworkMode: str
    IPv4Addresses: List[str]

    # doing this because AWS sends us keys we super don't care about
    def __init__(self, **kwargs):  # noqa: D107
        keys = set([field.name for field in fields(self)])
        for k, v in kwargs.items():
            if k in keys:
                setattr(self, k, v)


@dataclass(init=False, order=True)
class ECSContainerMetadata:
    """ECS container metadata."""

    DockerId: str
    Name: str
    DockerName: str
    Image: str
    ImageID: str
    Labels: Dict[str, str]
    DesiredStatus: str
    KnownStatus: str
    Limits: Dict[str, int]
    CreatedAt: str
    StartedAt: str
    Type: str
    Networks: List[ECSNetworkMetadata]

    # doing this because AWS sends us keys we super don't care about
    def __init__(self, **kwargs):  # noqa: D107
        keys = set([field.name for field in fields(self)])
        for k, v in kwargs.items():
            if k == "Networks":
                setattr(self, k, [ECSNetworkMetadata(**network) for network in v])
            elif k in keys:
                setattr(self, k, v)


@dataclass(init=False, order=True)
class ECSTaskMetadata:
    """ECS task metadata response object."""

    Cluster: str
    TaskARN: str
    Family: str
    Revision: str
    DesiredStatus: str
    KnownStatus: str
    Containers: List[ECSContainerMetadata]
    PullStartedAt: str
    PullStoppedAt: str
    AvailabilityZone: str

    # doing this because AWS sends us keys we super don't care about
    def __init__(self, **kwargs):  # noqa: D107
        keys = set([field.name for field in fields(self)])
        for k, v in kwargs.items():
            if k == "Containers":
                setattr(self, k, [ECSContainerMetadata(**container) for container in v])
            elif k in keys:
                setattr(self, k, v)



"""Main module."""

import hashlib
import json
import logging
import os
from typing import Dict, Generator, List

import boto3
import requests

from ecs_dns import METADATA_URL
from ecs_dns.models import ECSTaskMetadata

log = logging.getLogger("ecs-dns")
log.setLevel(os.getenv("LOG_LEVEL", "DEBUG"))


def find_private_ips(container_name=None) -> Generator[str, None, None]:
    """Invoke the metadata endpoint to get the private IPs for all containers in a running task."""
    raw_metadata = requests.get(METADATA_URL).text
    log.debug(f"Got metadata: {raw_metadata}")

    metadata = ECSTaskMetadata(**json.loads(raw_metadata))
    for container in metadata.Containers:
        if container_name and container_name != container.Name:
            continue
        for network in container.Networks:
            yield from network.IPv4Addresses


def _get_network_info(private_ip: str, result_key: str) -> Generator[str, None, None]:
    """Accept a private IP and yield the `result_key` value from each NetworkInterface's Association object."""
    client = boto3.client("ec2")
    filters = [{"Name": "addresses.private-ip-address", "Values": [private_ip]}]
    response = client.describe_network_interfaces(Filters=filters)
    log.debug(f"_get_network_info response: {response}")
    for interface in response["NetworkInterfaces"]:
        yield interface["Association"][result_key]


def find_public_dns_names(private_ip: str) -> List[str]:
    """Find the public DNS names associated with a private IP."""
    return list(_get_network_info(private_ip, "PublicDnsName"))


def find_public_ips(private_ip: str) -> List[str]:
    """Find the public IPs associated with a private IP."""
    return list(_get_network_info(private_ip, "PublicIp"))


def create_dns_record(public_ip: str, dns_name: str, healthcheck_id: str):
    """Create a Route53 multi-value A record with healthcheck."""
    client = boto3.client("route53")

    # set identifiers are used in multi-value record sets to uniquely id a record
    set_identifier = hashlib.md5(f"{public_ip}/{dns_name}")

    record_set = {
        "Name": dns_name,
        "Type": "A",
        "SetIdentifier": set_identifier,
        "MultiValueAnswer": True,
        # when using a health check, AWS recommends a TTL of 60 seconds or less.
        "TTL": 30,
        "ResourceRecords": [
            {"Value": public_ip},
        ],
        "HealthCheckId": healthcheck_id,
    }

    # get the domain's hosted zone id
    zone_name = ".".join(dns_name.split(".")[-2:])
    resp = client.list_hosted_zones_by_name(DNSName=zone_name)
    zone_id = resp["HostedZones"][0]["Id"].split("/")[-1]

    resp = client.change_resource_record_sets(
        HostedZoneId=zone_id,
        ChangeBatch={
            "Comment": "Created on container startup by ecs_dns.",
            "Changes": [{"Action": "UPSERT", "ResourceRecordSet": record_set}],
        },
    )


def create_health_check(
    public_ip: str,
    dns_name: str,
    port: int = 80,
    protocol: str = "HTTP",
    path: str = "/",
    interval: int = 30,
    failure_threshold: int = 3,
) -> str:
    """Create a Route53 healthcheck and return its ID.

    Args:
        public_ip: str
        dns_name: str
        port: int = 80
        protocol: str = "HTTP" - Can be 'HTTP'|'HTTPS'|'HTTP_STR_MATCH'|'HTTPS_STR_MATCH'|'TCP'|'CALCULATED'|'CLOUDWATCH_METRIC'|'RECOVERY_CONTROL'
        path: str = "/" - Ignored if type does not match HTTP*
        interval: int = 30
        failure_threshold: int = 3
    """
    client = boto3.client("route53")
    # caller reference is a unique string that identifies the request and that allows you to retry a failed
    # CreateHealthCheck request without the risk of creating two identical health checks
    caller_reference = hashlib.md5(f"{public_ip}/{dns_name}")
    response = client.create_health_check(
        CallerReference=caller_reference,
        HealthCheckConfig={
            "IPAddress": public_ip,
            "Port": port,
            "Type": protocol,
            "ResourcePath": path,
            "RequestInterval": interval,
            "FailureThreshold": failure_threshold,
            "Regions": ["us-east-1"],
        },
    )
    return response["HealthCheck"]["Id"]



from pathlib import Path
import datetime
from dateutil.tz import tzutc


def read_mock_metadata() -> str:
    path = Path(__file__).resolve().parent / "mock_metadata.json"
    with open(path, "r") as fh:
        return fh.read()


MOCK_DNI_PRIVATE_IP = "172.31.10.10"
MOCK_DNI_PUBLIC_IP = "50.200.80.30"
MOCK_DNI_PUBLIC_DNS = "ec2-50-200-80-30.compute-1.amazonaws.com"
MOCK_DNI_RESPONSE = {
    "NetworkInterfaces": [
        {
            "Association": {
                "IpOwnerId": "amazon",
                "PublicDnsName": MOCK_DNI_PUBLIC_DNS,
                "PublicIp": MOCK_DNI_PUBLIC_IP,
            },
            "Attachment": {
                "AttachTime": datetime.datetime(2021, 8, 1, 1, 1, 1, tzinfo=tzutc()),
                "AttachmentId": "eni-attach-123456789abcdefgh",
                "DeleteOnTermination": False,
                "DeviceIndex": 1,
                "NetworkCardIndex": 0,
                "InstanceOwnerId": "111111111111",
                "Status": "attached",
            },
            "AvailabilityZone": "us-east-1c",
            "Description": "arn:aws:ecs:us-east-1:111111111111:attachment/11111111-1234-1234-1234-111111111111",
            "Groups": [
                {"GroupName": "fake-security-group", "GroupId": "sg-123456789abcdefgh"}
            ],
            "InterfaceType": "interface",
            "Ipv6Addresses": [],
            "MacAddress": "12:34:56:78:90:12",
            "NetworkInterfaceId": "eni-123456789abcdefgh",
            "OwnerId": "111111111111",
            "PrivateDnsName": f"ip-{MOCK_DNI_PRIVATE_IP.replace('.', '-')}.ec2.internal",
            "PrivateIpAddress": MOCK_DNI_PRIVATE_IP,
            "PrivateIpAddresses": [
                {
                    "Association": {
                        "IpOwnerId": "amazon",
                        "PublicDnsName": MOCK_DNI_PUBLIC_DNS,
                        "PublicIp": MOCK_DNI_PUBLIC_IP,
                    },
                    "Primary": True,
                    "PrivateDnsName": f"ip-{MOCK_DNI_PRIVATE_IP.replace('.', '-')}.ec2.internal",
                    "PrivateIpAddress": MOCK_DNI_PRIVATE_IP,
                }
            ],
            "RequesterId": "123456789123",
            "RequesterManaged": True,
            "SourceDestCheck": True,
            "Status": "in-use",
            "SubnetId": "subnet-11111111",
            "TagSet": [
                {"Key": "aws:ecs:clusterName", "Value": "my-cluster"},
                {"Key": "aws:ecs:serviceName", "Value": "my-service"},
            ],
            "VpcId": "vpc-11111111",
        }
    ],
    "ResponseMetadata": {
        "RequestId": "6a870193-45a3-4a6b-a892-a1ec5caa6031",
        "HTTPStatusCode": 200,
        "HTTPHeaders": {
            "x-amzn-requestid": "6a870193-45a3-4a6b-a892-a1ec5caa6031",
            "cache-control": "no-cache, no-store",
            "strict-transport-security": "max-age=31536000; includeSubDomains",
            "content-type": "text/xml;charset=UTF-8",
            "content-length": "3112",
            "vary": "accept-encoding",
            "date": "Tue, 24 Aug 2021 20:35:22 GMT",
            "server": "AmazonEC2",
        },
        "RetryAttempts": 0,
    },
}



"""Console script for ECS CNAME."""
import sys

import click

from ecs_dns import (
    create_health_check,
    create_dns_record,
    find_private_ips,
    find_public_ips,
)


@click.command()
@click.argument("dns_name")
@click.option(
    "-c",
    "--container-name",
    default=None,
    type=click.STRING,
    help="Container name to create a record for.",
)
@click.option(
    "-p",
    "--port",
    default=80,
    type=click.INT,
    help="Port for the healthcheck to query.",
)
@click.option(
    "--protocol",
    default="HTTP",
    type=click.STRING,
    help="Protocol for the healthcheck to use.",
)
def cli(dns_name: str, container_name: str, port: int, protocol: str):
    """Automatically determine a container's public IP and create a multi-value A record and healthcheck for it.

    If multiple containers with public IPs are found, or if one container has multiple
    public IPs, only the first will be used.
    """
    public_ip_groups = [find_public_ips(ip) for ip in find_private_ips(container_name)]
    public_ips = list(set([i for group in public_ip_groups for i in group]))
    ip = public_ips[0]
    healthcheck = create_health_check(ip, dns_name, port, protocol)
    create_dns_record(ip, dns_name, healthcheck)


if __name__ == "__main__":
    sys.exit(cli())  # pragma: no cover



# -*- coding: utf-8 -*-

"""ECS CNAME: Creates a Route 53 CNAME for ECS tasks which are assigned a public IP.."""

__author__ = """Shaun Martin"""
__email__ = "inhumantsar@protonmail.com"
__version__ = "0.1.7"

METADATA_URL = "http://169.254.170.2/v2/metadata/"

from . import models
from ._ecs_dns import (
    create_health_check,
    create_dns_record,
    find_private_ips,
    find_public_dns_names,
    find_public_ips,
)


__all__ = [
    "models",
    "METADATA_URL",
    "find_private_ips",
    "find_public_dns_names",
    "find_public_ips",
    "create_health_check",
    "create_dns_record",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `ecs_dns` package."""

import pytest
from click.testing import CliRunner

from ecs_dns import _cli


def test_command_line_interface():
    """Test the CLI."""
    runner = CliRunner()

    # test no args
    result = runner.invoke(_cli.cli)
    assert result.exit_code == 2
    assert "Error: Missing argument 'DNS_NAME'" in result.output

    # test help command
    help_result = runner.invoke(_cli.cli, ["--help"])
    assert help_result.exit_code == 0
    expected_args = [
        "-c, --container-name TEXT",
        "Container name to create a record for.",
        "-p, --port INTEGER",
        "Port for the healthcheck to query.",
        "--protocol TEXT",
        "Protocol for the healthcheck to use.",
        "--help",
        "Show this message and exit.",
    ]
    assert all(arg in help_result.output for arg in expected_args)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `ecs_dns` package."""

import pytest
import os

from pprint import pprint
from pathlib import Path

import boto3
import responses

from moto import mock_ec2

from ecs_dns import find_private_ips, find_public_ips, METADATA_URL
from ecs_dns._ecs_dns import _get_network_info
from tests.mock_data import read_mock_metadata, MOCK_DNI_PRIVATE_IP


@pytest.fixture(scope="function")
def aws_credentials():
    """Mocked AWS Credentials for moto."""
    os.environ["AWS_ACCESS_KEY_ID"] = "testing"
    os.environ["AWS_SECRET_ACCESS_KEY"] = "testing"
    os.environ["AWS_SECURITY_TOKEN"] = "testing"
    os.environ["AWS_SESSION_TOKEN"] = "testing"


@responses.activate
def test_metadata_parse():
    """Test metadata parsing using mock data."""
    responses.add(
        responses.GET,
        METADATA_URL,
        body=read_mock_metadata(),
        status=200,
    )

    assert list(find_private_ips()) == ["10.0.2.106", "10.0.2.106"]


@mock_ec2
def test_get_network_info(aws_credentials):
    """Test parsing boto3 response."""
    # using a dummy ec2 instance since the describe network interfaces call
    # works the same way with fargate
    client = boto3.client("ec2", region_name="us-east-1")
    vpc_id = client.describe_vpcs()["Vpcs"][0]["VpcId"]
    subnets = client.describe_subnets(Filters=[{"Name": "vpc-id", "Values": [vpc_id]}])
    subnet_id = subnets["Subnets"][0]["SubnetId"]
    client.modify_vpc_attribute(
        EnableDnsSupport={"Value": True},
        VpcId=vpc_id,
    )
    client.modify_vpc_attribute(
        EnableDnsHostnames={"Value": True},
        VpcId=vpc_id,
    )
    resp = client.create_network_interface(
        PrivateIpAddress=MOCK_DNI_PRIVATE_IP, SubnetId=subnet_id
    )
    eni_id = resp["NetworkInterface"]["NetworkInterfaceId"]
    resp = client.describe_network_interfaces(
        Filters=[{"Name": "network-interface-id", "Values": [eni_id]}]
    )
    association = resp["NetworkInterfaces"][0]["Association"]
    expected_ip = association["PublicIp"]
    expected_dns = association["PublicDnsName"]
    found_ip = list(_get_network_info(MOCK_DNI_PRIVATE_IP, "PublicIp"))[0]
    found_dns = list(_get_network_info(MOCK_DNI_PRIVATE_IP, "PublicDnsName"))[0]
    assert expected_ip == found_ip
    assert expected_dns == found_dns
